import java.awt.image.*;
import javax.imageio.*;
import java.io.*;
import java.util.*;

public class Main {

  public static final int SQRT_SAMPLES = 1;
  public static final int INVERSE_SAMPLES = 1 / (SQRT_SAMPLES * SQRT_SAMPLES);
  public static final int WIDTH = 640;
  public static final int HEIGHT = 512;
  public static final String IMAGE_TYPE = "png";
  public static final String OUTPUT_FILE = "output." + IMAGE_TYPE;
  public static final int NOISE_SIZE = 1 << 8;
  public static final int NOISE_MASK = NOISE_SIZE - 1;
  public static final long NOISE_SEED = 0xDEADBEEFCAFEBABEL;
  public static final double GAIN = 0.5;
  public static final double LACUNARITY = 2.5;
  public static final double FBM_SCALE = 0.02;
  public static final int FBM_ITERATIONS = 16;

  public static final int RED = 0;
  public static final int GREEN = 1;
  public static final int BLUE = 2;

  public static final double GAMMA = 2.2;
  public static final double INVERSE_GAMMA = 1.0 / GAMMA;

  public static final long SECOND_MILLIS = 1000L;
  public static final long MINUTE_MILLIS = 60 * SECOND_MILLIS;
  public static final long HOUR_MILLIS = 60 * MINUTE_MILLIS;

  private double[][][] pixels = new double[HEIGHT][WIDTH][3];

  private double[][] noise = new double[NOISE_SIZE][NOISE_SIZE];

  private void launch() throws Throwable {

    initializeNoise();

    for(int y = 0; y < HEIGHT; y++) {
      for(int x = 0; x < WIDTH; x++) {
        double intensity = fractionalBrownianMotion(x, y);
        for(int i = 0; i < 3; i++) {
          pixels[y][x][i] = intensity;
        }
      }
    }

    saveImage();

  }

  private double fractionalBrownianMotion(double x, double y) {

    double gain = 1.0;
    double lacunarity = FBM_SCALE;
    double sum = 0.0;

    for(int i = 0; i < FBM_ITERATIONS;
        i++, gain *= GAIN, lacunarity *= LACUNARITY) {
      sum += gain * noiseFunction(x * lacunarity, y * lacunarity);
    }

    return sum;
  }

  private double noiseFunction(double x, double y) {

    int floorX = (int)Math.floor(x);
    int floorY = (int)Math.floor(y);

    double percentX = x - floorX;
    double percentY = y - floorY;

    int x1 = floorX & NOISE_MASK;
    int y1 = floorY & NOISE_MASK;
    int x2 = (x1 + 1) & NOISE_MASK;
    int y2 = (y1 + 1) & NOISE_MASK;

    double n11 = noise[y1][x1];
    double n21 = noise[y1][x2];
    double n12 = noise[y2][x1];
    double n22 = noise[y2][x2];

    double a = n11 + (n12 - n11) * percentY;
    double b = n21 + (n22 - n21) * percentY;

    return a + (b - a) * percentX;
  }

  private void initializeNoise() {
    Random random = new Random(NOISE_SEED);
    for(int i = 0; i < NOISE_SIZE; i++) {
      for(int j = 0; j < NOISE_SIZE; j++) {
        noise[i][j] = random.nextDouble();
      }
    }
  }

  private void saveImage() throws Throwable {
    int[] data = new int[WIDTH * HEIGHT];

    for(int y = 0, k = 0; y < HEIGHT; y++) {
      for(int x = 0; x < WIDTH; x++, k++) {
        int value = 0xFF;
        for(int i = 0; i < 3; i++) {
          int intensity = (int)Math.round(255
              * Math.pow(pixels[y][x][i] * INVERSE_SAMPLES, INVERSE_GAMMA));
          if (intensity < 0) {
            intensity = 0;
          } else if (intensity > 255) {
            intensity = 255;
          }
          value <<= 8;
          value |= intensity;
        }
        data[k] = value;
      }
    }

    BufferedImage image = new BufferedImage(
        WIDTH, HEIGHT, BufferedImage.TYPE_INT_ARGB_PRE);
    image.setRGB(0, 0, WIDTH, HEIGHT, data, 0, WIDTH);
    ImageIO.write(image, IMAGE_TYPE, new File(OUTPUT_FILE));
  }

  public static void main(String... args) throws Throwable {

    long startTime = System.currentTimeMillis();

    Main main = new Main();
    main.launch();

    long interval = System.currentTimeMillis() - startTime;
    long hours = interval / HOUR_MILLIS;
    interval %= HOUR_MILLIS;
    long minutes = interval / MINUTE_MILLIS;
    interval %= MINUTE_MILLIS;
    long seconds = interval / SECOND_MILLIS;
    interval %= SECOND_MILLIS;
    System.out.format("%d hour%s, %d minute%s, %d second%s, %d millisecond%s%n",
        hours, hours == 1 ? "" : "s",
        minutes, minutes == 1 ? "" : "s",
        seconds, seconds == 1 ? "" : "s",
        interval, interval == 1 ? "" : "s");
  }

}
